#!/opt/anaconda1anaconda2anaconda3\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'glueviz==0.8.1','console_scripts','glue-deps'
__requires__ = 'glueviz==0.8.1'
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.exit(
        load_entry_point('glueviz==0.8.1', 'console_scripts', 'glue-deps')()
    )
